from warnings import warn
import heapq

class Node:
    """
    A node class for A* Pathfinding
    """

    def __init__(self, parent=None, position=None):
        self.parent = parent
        self.position = position

        self.g = 0
        self.h = 0
        self.f = 0

    def __eq__(self, other):
        return self.position == other.position
    
    def __repr__(self):
      return f"{self.position} - g: {self.g} h: {self.h} f: {self.f}"

    # defining less than for purposes of heap queue
    def __lt__(self, other):
      return self.f < other.f
    
    # defining greater than for purposes of heap queue
    def __gt__(self, other):
      return self.f > other.f

def return_path(current_node):
    path = []
    current = current_node
    while current is not None:
        path.append(current.position)
        current = current.parent
    return path[::-1]  # Return reversed path


def astar(maze, start, end, allow_diagonal_movement = True):
    """
    Returns a list of tuples as a path from the given start to the given end in the given maze
    :param maze:
    :param start:
    :param end:
    :return:
    """

    # Create start and end node
    start_node = Node(None, start)
    start_node.g = start_node.h = start_node.f = 0
    end_node = Node(None, end)
    end_node.g = end_node.h = end_node.f = 0

    # Initialize both open and closed list
    open_list = []
    closed_list = []

    # Heapify the open_list and Add the start node
    heapq.heapify(open_list) 
    heapq.heappush(open_list, start_node)

    # Adding a stop condition
    outer_iterations = 0
    max_iterations = (len(maze[0]) * len(maze) // 2)

    # what squares do we search
    adjacent_squares = ((0, -1), (0, 1), (-1, 0), (1, 0),)
    if allow_diagonal_movement:
        adjacent_squares = ((0, -1), (0, 1), (-1, 0), (1, 0), (-1, -1), (-1, 1), (1, -1), (1, 1),)

    # Loop until you find the end
    while len(open_list) > 0:
        outer_iterations += 1

        if outer_iterations > max_iterations:
          # if we hit this point return the path such as it is
          # it will not contain the destination
          warn("giving up on pathfinding too many iterations")
          return return_path(current_node)       
        
        # Get the current node
        current_node = heapq.heappop(open_list)
        closed_list.append(current_node)

        # Found the goal
        if current_node == end_node:
            return return_path(current_node)

        # Generate children
        children = []
        
        for new_position in adjacent_squares: # Adjacent squares

            # Get node position
            node_position = (current_node.position[0] + new_position[0], current_node.position[1] + new_position[1])

            # Make sure within range
            if node_position[0] > (len(maze) - 1) or node_position[0] < 0 or node_position[1] > (len(maze[len(maze)-1]) -1) or node_position[1] < 0:
                continue

            # Make sure walkable terrain
            if maze[node_position[0]][node_position[1]] != 0:
                continue

            # Create new node
            new_node = Node(current_node, node_position)

            # Append
            children.append(new_node)

        # Loop through children
        for child in children:
            # Child is on the closed list
            if len([closed_child for closed_child in closed_list if closed_child == child]) > 0:
                continue

            # Create the f, g, and h values
            child.g = current_node.g + 1
            child.h = ((child.position[0] - end_node.position[0]) ** 2) + ((child.position[1] - end_node.position[1]) ** 2)
            child.f = child.g + child.h

            # Child is already in the open list
            if len([open_node for open_node in open_list if child.position == open_node.position and child.g > open_node.g]) > 0:
                continue

            # Add the child to the open list
            heapq.heappush(open_list, child)

    warn("Couldn't get a path to destination")
    return None


def points_start(maze):
    p_t = maze[0]
    p_end = []
    for i in range(len(p_t)):
        if(maze[i][0] == 0):
            p_end.append((i, 0))
    return p_end


def points_end(maze):
    p_t = maze[0]
    size = len(p_t)-1
    p_end = []
    for i in range(size+1):
        if(maze[i][size] == 0):
            p_end.append((i, size))
    return p_end



import numpy as np
import matplotlib.pyplot as plt
import matplotlib
from taugeometric.poresSE import points_start
from taugeometric.poresSE import points_end
from taugeometric.astar import astar
from scipy.spatial import distance as ds

def metric_euclidian(points):
    p_s=points[0]
    distance=0
    for i in range(1,len(points)):
        p_e=points[i]
        distance+=ds.euclidean(p_s,p_e)
        p_s=points[i]
    return distance


    
def show_streamlines(maze,paths):
    plt.imshow(maze,cmap="gray_r")
    for i in paths:
        for j in i:
            maze[j[0]][j[1]]=2
    colors = 'white dimgray black '.split()
    cmap = matplotlib.colors.ListedColormap(colors, name='colors', N=None)
    plt.imshow(maze,cmap=cmap)
    
def tortuosity_geometric_2d(maze,showStreamline=False):
    line_maze_h=maze.shape[1]-1
    p_s = points_start(maze)
    p_e = points_end(maze)

    f_data = []
    path_show=[]
    for i in p_s:
        selection_path=[]
        
        pahts=[]
        for j in p_e:
            data = astar(maze, i, j)
            if(data != None):
                selection_path.append(metric_euclidian(data))
                pahts.append(data)
                
        minimum=np.array(selection_path).min()
        position=selection_path.index(minimum)
        path_show.append(pahts[position])
        f_data.append(minimum)
    if(showStreamline):
        show_streamlines(maze,path_show)
    t_g=np.mean(f_data)/(line_maze_h)
    return t_g


    



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="taugeometric",
    version="1.0.0",
    author="Mayken Espinoza -  Joseph Avila - Javier Pagalo ",
    author_email="josdavil@espol.edu.ec",
    description="TauGeometric is a set of tools to get the geometric tortuosity from 2D images of porous materials",
    long_description_content_type="text/markdown",
    url="https://github.com/eljosephavila123/taugeomectic",
    download_url='https://pypi.org/project/taugeometric/',
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    
    python_requires='>=3.6',
)
install_requires = [
    'matplotlib==3.3.3',
     ' numpy==1.19.5',
      'porespy==1.3.1'

]


